==> delegate8.7.8.tar.gz: delegate8.7.8/rary/pam.c <==
/*////////////////////////////////////////////////////////////////////////
Copyright (c) 2003 National Institute of Advanced Industrial Science and Technology (AIST)

Permission to use this material for evaluation, copy this material for
your own use, and distribute the copies via publically accessible on-line
media, without fee, is hereby granted provided that the above copyright
notice and this permission notice appear in all copies.
AIST MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.

/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	pam.c (PAM interface)
Author:		Yutaka Sato <ysato@delegate.org>
Description:
History:
	030814	created
////////////////////////////////////////////////////////////////////////*/
#include <stdio.h>
#include "ystring.h"

typedef struct {
	char	p_request[128];
	char	p_method[64];
	char	p_service[64];
	char	p_user[64];
	char	p_pass[64];
} PamReq;

#define PAM_OK		200
#define PAM_UNAUTH	401
#define PAM_FORBIDDEN	403

#define DEFAULT_PAMbaseurl "/-/pam/"
extern char *PAMbaseurl;
extern char *PAMurl;

/*
 * PAM server
 */
static get_pamreq(qfp,PQ)
	FILE *qfp;
	PamReq *PQ;
{	int gotop = 0;
	char line[128],serv[64],atype[32],xauth[128],auth[128],*dp;

	PQ->p_request[0] = 0;
	PQ->p_method[0] = 0;
	PQ->p_service[0] = 0;
	PQ->p_user[0] = 0;
	PQ->p_pass[0] = 0;

	while( fgets(line,sizeof(line),qfp) != NULL ){
		if( *line == '\r' || *line == '\n' )
			break;
		if( gotop == 0 ){
			lineScan(line,PQ->p_request);
			dp = wordScan(line,PQ->p_method);
			wordScan(dp,serv);
			if( strncmp(serv,PAMbaseurl,strlen(PAMbaseurl)) == 0 )
				ovstrcpy(serv,serv+strlen(PAMbaseurl));
			dp = wordscanY(serv,
				PQ->p_service,sizeof(PQ->p_service),"^/");
			if( *dp == '/' )
				wordScan(dp+1,PQ->p_method);
			gotop = 1;
		}else
		if( strncasecmp(line,"Authorization:",14) == 0 ){
			dp = wordScan(line+14,atype);
			dp = wordScan(dp,xauth);
			str_from64(xauth,strlen(xauth),auth,sizeof(auth));
			scan_field1(auth,PQ->p_user,sizeof(PQ->p_user),
				PQ->p_pass,sizeof(PQ->p_pass));
		}
	}
}
static put_pamreq(qfp,host,port,method,service,user,pass)
	FILE *qfp;
	char *method,*service,*user,*pass;
{	char auth[128],xauth[128];

	sprintf(auth,"%s:%s",user,pass);
	str_to64(auth,strlen(auth),xauth,sizeof(xauth),1);

	if( PAMurl )
		fprintf(qfp,"GET %s HTTP/1.0\r\n",PAMurl);
	else	fprintf(qfp,"GET %s%s/%s HTTP/1.0\r\n",
			0<port?PAMbaseurl:"",service,method);
	if( 0 < port )
		fprintf(qfp,"Host: %s:%d\r\n",host,port);
	fprintf(qfp,"Authorization: Basic %s\r\n",xauth);
	fprintf(qfp,"\r\n");
	fflush(qfp);
}
static get_pamresp(rfp,rstat,rsize)
	FILE *rfp;
	char *rstat;
{	int ok,rcode;

	ok = -1;
	if( fgets(rstat,rsize,rfp) != NULL ){
		rcode = 0;
		sscanf(rstat,"%*s %d",&rcode);
		if( rcode == PAM_OK )
			ok = 1;
		else	ok = 0;
	}
	return ok;
}
static put_pamresp(rfp,ok)
	FILE *rfp;
{	int rcode;

	if( 0 < ok )
		rcode = PAM_OK;
	else
	if( ok < 0 )
		rcode = PAM_FORBIDDEN;
	else	rcode = PAM_UNAUTH;
	fprintf(rfp,"HTTP/1.0 %d\r\n",rcode);
	fprintf(rfp,"\r\n");
	fprintf(rfp,"PAM-Expire: none\r\n");
	fflush(rfp);
	return rcode;
}

#ifndef PAM_BY_MAIN_ONLY

char *PAMbaseurl = DEFAULT_PAMbaseurl;
char *PAMurl;

/*
#include <security/pam_appl.h>
*/
struct pam_message { int msg_style; const char *msg; };
struct pam_response { char *resp; int resp_retcode; };
struct pam_conv { int (*conv)(); void *appdata_ptr; };
#define PAM_PROMPT_ECHO_OFF 1
#define PAM_SUCCESS 0

static char *gpass;
static login_conv(nmsg,pmsg,resp,appdata)
	struct pam_message **pmsg;
	struct pam_response **resp;
	char *appdata;
{	int mi;

	for( mi = 0; mi < nmsg; mi++ ){
		switch( pmsg[mi]->msg_style ){
		case PAM_PROMPT_ECHO_OFF:
			resp[mi] = calloc(1,sizeof(struct pam_response));
			resp[mi]->resp = strdup(gpass);
			break;
		}
	}
	return PAM_SUCCESS;
}
static struct pam_conv pam_conv = {login_conv, NULL};

pam_auth1(service,user,pass)
	char *service,*user,*pass;
{	int status;
	void *pamh;

	if( pam_start(service,user,&pam_conv,&pamh) != PAM_SUCCESS ){
		syslog_ERROR("PAM: cannot start %s [%s]\n",service,user);
		return -1;
	}
	gpass = pass;
	status = pam_authenticate(pamh,0);
	pam_end(pamh,PAM_SUCCESS);
	syslog_ERROR("## pam_authenticate [%s][%s] = %d\n",service,user,status);

	if( status != PAM_SUCCESS ){
		return 0;
	}
	return 1;
}

pam_server(ac,av)
	char *av[];
{	int ok;
	PamReq PQ;

	get_pamreq(stdin,&PQ);
	ok = pam_auth1(PQ.p_service,PQ.p_user,PQ.p_pass);
	put_pamresp(stdout,ok);
	return ok;
}

#else /* PAM_BY_MAIN_ONLY */

static char *ext_pam = "dgpam";
static char *pam_path;
static forkPAMserver(sv)
	int sv[];
{	int ac,pid;
	char *av[8],path[1024];
	int qpipe[2],rpipe[2];

	ac = 0;
	av[ac++] = "pam";
	av[ac] = 0;

	if( pam_path == 0 ){
		strcpy(path,ext_pam);
		if( fullpathSUCOM(ext_pam,"r",path) == 0 ){
			syslog_ERROR("## command not found: %s\n",ext_pam);
			return -1;
		}
		syslog_ERROR("## dgpam = %s\n",path);
		pam_path = stralloc(path);
	}

	pipe(qpipe);
	pipe(rpipe);
	if( (pid = fork()) == 0 ){
		close(qpipe[1]); dup2(qpipe[0],0); close(qpipe[0]);
		close(rpipe[0]); dup2(rpipe[1],1); close(rpipe[1]);
		Execvp("Pam",pam_path,av);
		exit(-1);
	}
	close(qpipe[0]);
	close(rpipe[1]);

	sv[0] = rpipe[0];
	sv[1] = qpipe[1];
	return pid;
}

static openPAMserver(ctx,host,port,sv)
	void *ctx;
	char *host;
	int sv[];
{	int sock,pid;
	char local[256],remote[256];

	if( 0 < port ){
		sprintf(remote,"%s:%d",host,port);
		sock = VSocket(ctx,"CNCT/PAM",-1,local,remote,"proto=pam,FSV");
		if( 0 <= sock ){
			sv[0] = sv[1] = sock;
			return 0;
		}
		return -1;
	}

	if( 0 < (pid = forkPAMserver(sv)) )
		return pid;

	return -1;
}

static pam_auth1x(ctx,host,port,service,user,pass)
	void *ctx;
	char *host,*service,*user,*pass;
{	int rcode,ok;
	int pamsv[2];
	FILE *qfp,*rfp;
	char rstat[128];
	int pid;

	pid = openPAMserver(ctx,host,port,pamsv);
	if( pid < 0 ){
		return -1;
	}

	qfp = fdopen(pamsv[1],"w");
	rfp = fdopen(pamsv[0],"r");
	put_pamreq(qfp,host,port,"auth",service,user,pass);

	ok = get_pamresp(rfp,rstat,sizeof(rstat));
	syslog_ERROR("## dgpam -a %s %s = %s",service,user,rstat);

	fclose(qfp);
	fclose(rfp);
	if( 0 < pid )
		wait(0);
	return ok;
}

pam_service(ctx,forbidden,req,user,stcodep)
	void *ctx;
	char *req,*user;
	int *stcodep;
{	int ok;
	PamReq PQ;

	if( forbidden ){
		ok = -1;
		bzero(&PQ,sizeof(PamReq));
		*stcodep = put_pamresp(stdout,ok);
	}else{
		get_pamreq(stdin,&PQ);
		ok = pam_auth1x(ctx,"-",0,PQ.p_service,PQ.p_user,PQ.p_pass);
		*stcodep = put_pamresp(stdout,ok);
	}
	strcpy(req,PQ.p_request);
	strcpy(user,PQ.p_user);
	return ok;
}

pam_checkPasswd(ctx,host,port,service,user,pass)
	void *ctx;
	char *host,*service,*user,*pass;
{	int ok;
	int euid;

	ok = 0;
	euid = geteuid();

	/* ?? under non-superuser privilege,
	 * if password is correct for the user who own the porcess,
	 * PAM seems to return OK regardlessly of the username ??
	 */
	if( port == 0 )
	if( getUserId(user) == euid || euid == 0 ){
		ok = pam_auth1(service,user,pass);
	}
	if( port != 0 || !ok && euid != 0 && INHERENT_fork() ){
		ok = pam_auth1x(ctx,host,port,service,user,pass);
	}
	return ok;
}

#endif /* PAM_BY_MAIN_ONLY */

==> delegate8.9.0.tar.gz: delegate8.9.0/rary/pam.c <==
/*////////////////////////////////////////////////////////////////////////
Copyright (c) 2003 National Institute of Advanced Industrial Science and Technology (AIST)

Permission to use this material for evaluation, copy this material for
your own use, and distribute the copies via publically accessible on-line
media, without fee, is hereby granted provided that the above copyright
notice and this permission notice appear in all copies.
AIST MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.

/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	pam.c (PAM interface)
Author:		Yutaka Sato <ysato@delegate.org>
Description:
History:
	030814	created
////////////////////////////////////////////////////////////////////////*/
#include <stdio.h>
#include "ystring.h"

typedef struct {
	char	p_request[128];
	char	p_method[64];
	char	p_service[64];
	char	p_user[64];
	char	p_pass[64];
} PamReq;

#define PAM_OK		200
#define PAM_UNAUTH	401
#define PAM_FORBIDDEN	403

#define DEFAULT_PAMbaseurl "/-/pam/"
extern char *PAMbaseurl;
extern char *PAMurl;

/*
 * PAM server
 */
static get_pamreq(qfp,PQ)
	FILE *qfp;
	PamReq *PQ;
{	int gotop = 0;
	char line[128],serv[64],atype[32],xauth[128],auth[128],*dp;

	PQ->p_request[0] = 0;
	PQ->p_method[0] = 0;
	PQ->p_service[0] = 0;
	PQ->p_user[0] = 0;
	PQ->p_pass[0] = 0;

	while( fgets(line,sizeof(line),qfp) != NULL ){
		if( *line == '\r' || *line == '\n' )
			break;
		if( gotop == 0 ){
			lineScan(line,PQ->p_request);
			dp = wordScan(line,PQ->p_method);
			wordScan(dp,serv);
			if( strncmp(serv,PAMbaseurl,strlen(PAMbaseurl)) == 0 )
				ovstrcpy(serv,serv+strlen(PAMbaseurl));
			dp = wordscanY(serv,
				PQ->p_service,sizeof(PQ->p_service),"^/");
			if( *dp == '/' )
				wordScan(dp+1,PQ->p_method);
			gotop = 1;
		}else
		if( strncasecmp(line,"Authorization:",14) == 0 ){
			dp = wordScan(line+14,atype);
			dp = wordScan(dp,xauth);
			str_from64(xauth,strlen(xauth),auth,sizeof(auth));
			scan_field1(auth,PQ->p_user,sizeof(PQ->p_user),
				PQ->p_pass,sizeof(PQ->p_pass));
		}
	}
}
static put_pamreq(qfp,host,port,method,service,user,pass)
	FILE *qfp;
	char *method,*service,*user,*pass;
{	char auth[128],xauth[128];

	sprintf(auth,"%s:%s",user,pass);
	str_to64(auth,strlen(auth),xauth,sizeof(xauth),1);

	if( PAMurl )
		fprintf(qfp,"GET %s HTTP/1.0\r\n",PAMurl);
	else	fprintf(qfp,"GET %s%s/%s HTTP/1.0\r\n",
			0<port?PAMbaseurl:"",service,method);
	if( 0 < port )
		fprintf(qfp,"Host: %s:%d\r\n",host,port);
	fprintf(qfp,"Authorization: Basic %s\r\n",xauth);
	fprintf(qfp,"\r\n");
	fflush(qfp);
}
static get_pamresp(rfp,rstat,rsize)
	FILE *rfp;
	char *rstat;
{	int ok,rcode;

	ok = -1;
	if( fgets(rstat,rsize,rfp) != NULL ){
		rcode = 0;
		sscanf(rstat,"%*s %d",&rcode);
		if( rcode == PAM_OK )
			ok = 1;
		else	ok = 0;
	}
	return ok;
}
static put_pamresp(rfp,ok)
	FILE *rfp;
{	int rcode;

	if( 0 < ok )
		rcode = PAM_OK;
	else
	if( ok < 0 )
		rcode = PAM_FORBIDDEN;
	else	rcode = PAM_UNAUTH;
	fprintf(rfp,"HTTP/1.0 %d\r\n",rcode);
	fprintf(rfp,"\r\n");
	fprintf(rfp,"PAM-Expire: none\r\n");
	fflush(rfp);
	return rcode;
}

#ifndef PAM_BY_MAIN_ONLY

char *PAMbaseurl = DEFAULT_PAMbaseurl;
char *PAMurl;

/*
#include <security/pam_appl.h>
*/
struct pam_message { int msg_style; char *msg; };
struct pam_response { char *resp; int resp_retcode; };
struct pam_conv { int (*conv)(); void *appdata_ptr; };
#define PAM_PROMPT_ECHO_OFF 1
#define PAM_SUCCESS 0

extern char *calloc();
static char *gpass;
static login_conv(nmsg,pmsg,resp,appdata)
	struct pam_message **pmsg;
	struct pam_response **resp;
	char *appdata;
{	int mi;

	for( mi = 0; mi < nmsg; mi++ ){
		switch( pmsg[mi]->msg_style ){
		case PAM_PROMPT_ECHO_OFF:
			resp[mi] = (struct pam_response*)calloc(1,sizeof(struct pam_response));
			resp[mi]->resp = strdup(gpass);
			break;
		}
	}
	return PAM_SUCCESS;
}
static struct pam_conv pam_conv = {login_conv, NULL};

pam_auth1(service,user,pass)
	char *service,*user,*pass;
{	int status;
	void *pamh;

	/*
	if( pam_start(service,user,&pam_conv,&pamh) != PAM_SUCCESS ){
		syslog_ERROR("PAM: cannot start %s [%s]\n",service,user);
	*/
	pamh = NULL;
	status = pam_start(service,user,&pam_conv,&pamh);
	if( status != PAM_SUCCESS ){
		syslog_ERROR("PAM: cannot start %s [%s] error=%d\n",
			service,user,status);
		return -1;
	}
	gpass = pass;
	status = pam_authenticate(pamh,0);
	pam_end(pamh,PAM_SUCCESS);
	syslog_ERROR("## pam_authenticate [%s][%s] = %d\n",service,user,status);

	if( status != PAM_SUCCESS ){
		return 0;
	}
	return 1;
}

pam_server(ac,av)
	char *av[];
{	int ok;
	PamReq PQ;

	get_pamreq(stdin,&PQ);
	ok = pam_auth1(PQ.p_service,PQ.p_user,PQ.p_pass);
	put_pamresp(stdout,ok);
	return ok;
}

#else /* PAM_BY_MAIN_ONLY */

static char *ext_pam = "dgpam";
static char *pam_path;
static forkPAMserver(sv)
	int sv[];
{	int ac,pid;
	char *av[8],path[1024];
	int qpipe[2],rpipe[2];

	ac = 0;
	av[ac++] = "pam";
	av[ac] = 0;

	if( pam_path == 0 ){
		strcpy(path,ext_pam);
		if( fullpathSUCOM(ext_pam,"r",path) == 0 ){
			syslog_ERROR("## command not found: %s\n",ext_pam);
			return -1;
		}
		syslog_ERROR("## dgpam = %s\n",path);
		pam_path = stralloc(path);
	}

	pipe(qpipe);
	pipe(rpipe);
	if( (pid = fork()) == 0 ){
		close(qpipe[1]); dup2(qpipe[0],0); close(qpipe[0]);
		close(rpipe[0]); dup2(rpipe[1],1); close(rpipe[1]);
		Execvp("Pam",pam_path,av);
		exit(-1);
	}
	close(qpipe[0]);
	close(rpipe[1]);

	sv[0] = rpipe[0];
	sv[1] = qpipe[1];
	return pid;
}

static openPAMserver(ctx,host,port,sv)
	void *ctx;
	char *host;
	int sv[];
{	int sock,pid;
	char local[256],remote[256];

	if( 0 < port ){
		sprintf(remote,"%s:%d",host,port);
		sock = VSocket(ctx,"CNCT/PAM",-1,local,remote,"proto=httpam,FSV");
		if( 0 <= sock ){
			sv[0] = sv[1] = sock;
			return 0;
		}
		return -1;
	}

	if( 0 < (pid = forkPAMserver(sv)) )
		return pid;

	return -1;
}

static pam_auth1x(ctx,host,port,service,user,pass)
	void *ctx;
	char *host,*service,*user,*pass;
{	int rcode,ok;
	int pamsv[2];
	FILE *qfp,*rfp;
	char rstat[128];
	int pid;

	pid = openPAMserver(ctx,host,port,pamsv);
	if( pid < 0 ){
		return -1;
	}

	qfp = fdopen(pamsv[1],"w");
	rfp = fdopen(pamsv[0],"r");
	put_pamreq(qfp,host,port,"auth",service,user,pass);

	ok = get_pamresp(rfp,rstat,sizeof(rstat));
	syslog_ERROR("## dgpam -a %s %s = %s",service,user,rstat);

	fclose(qfp);
	fclose(rfp);
	if( 0 < pid )
		wait(0);
	return ok;
}

pam_service(ctx,forbidden,req,user,stcodep)
	void *ctx;
	char *req,*user;
	int *stcodep;
{	int ok;
	PamReq PQ;

	if( forbidden ){
		ok = -1;
		bzero(&PQ,sizeof(PamReq));
		*stcodep = put_pamresp(stdout,ok);
	}else{
		get_pamreq(stdin,&PQ);
		ok = pam_auth1x(ctx,"-",0,PQ.p_service,PQ.p_user,PQ.p_pass);
		*stcodep = put_pamresp(stdout,ok);
	}
	strcpy(req,PQ.p_request);
	strcpy(user,PQ.p_user);
	return ok;
}

pam_checkPasswd(ctx,host,port,service,user,pass)
	void *ctx;
	char *host,*service,*user,*pass;
{	int ok;
	int euid;

	ok = 0;
	euid = geteuid();

	/* ?? under non-superuser privilege,
	 * if password is correct for the user who own the porcess,
	 * PAM seems to return OK regardlessly of the username ??
	 */
	if( port == 0 )
	if( getUserId(user) == euid || euid == 0 ){
		ok = pam_auth1(service,user,pass);
	}
	if( port != 0 || !ok && euid != 0 && INHERENT_fork() ){
		ok = pam_auth1x(ctx,host,port,service,user,pass);
	}
	return ok;
}

#endif /* PAM_BY_MAIN_ONLY */
